/* Copyright (c) 2007, Christopher Walton
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of 
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <cstdio>
#include "pluginterfaces/vst2.x/aeffectx.h"
#include "Upsampler2xFpu.h"
#include "Downsampler2xFpu.h"
#include "Upsampler2xSse.h"
#include "Downsampler2xSse.h"
#include "Upsampler2x3dnow.h"
#include "Downsampler2x3dnow.h"

#define NAME_PREFIX				"[OS] "
#define TITLE					"VST Oversampler"
#define WND_OFFSET				17

#define CLASSNAME_SELF			"VST Oversampler - Wrapper"
#define CLASSNAME_CLIENT		"VST Oversampler - Client"

enum {
	CPUTYPE_X87 = 0,
	CPUTYPE_3DNOW,
	CPUTYPE_SSE,

	MAX_PATH_SIZE = 1024,
	N_COEFFS_UP = 8,
	N_COEFFS_DOWN = 12,
};

typedef unsigned int uint;
typedef AEffect* VstProc(audioMasterCallback audioMaster);

void* operator new(size_t cbSize);
void operator delete(void* p);


VstIntPtr vst_audioMaster(AEffect* aeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt);
VstIntPtr vst_dispatcher(AEffect* aeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt);
void vst_processReplacing(AEffect* aeff, float** inputs, float** outputs, VstInt32 sampleFrames);
void vst_processAccumulating(AEffect* aeff, float** inputs, float** outputs, VstInt32 sampleFrames);
void vst_processDoubleReplacing(AEffect* aeff, double** inputs, double** outputs, VstInt32 sampleFrames);
float vst_getParameter(AEffect* aeff, VstInt32 index);
void vst_setParameter(AEffect* aeff, VstInt32 index, float value);

extern "C" {
	LRESULT CALLBACK wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
}

// only compiled in when LOGGING_ENABLED is defined
#ifdef LOGGING_ENABLED
extern FILE* g_logFile;
void ____toLog(char* file, int lineNumber, char* format, ...);
#define toLog(...) ____toLog(__FILE__, __LINE__, __VA_ARGS__)
#else // LOGGING_ENABLED
#define toLog(...)
#endif // LOGGING_ENABLED


class Wrapper
{
public:
	static Wrapper* createWrapper();

public:
	Wrapper();
	virtual ~Wrapper();

public:
	VstIntPtr vst_audioMaster(AEffect* originalAeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt);
	VstIntPtr vst_dispatcher(VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt);
	void vst_processReplacing(float** inputs, float** outputs, VstInt32 sampleFrames);
	void vst_processAccumulating(float** inputs, float** outputs, VstInt32 sampleFrames);
	void vst_processDoubleReplacing(double** inputs, double** outputs, VstInt32 sampleFrames);
	float vst_getParameter(VstInt32 index);
	void vst_setParameter(VstInt32 index, float value);
	
protected:
	virtual void upsample(float** inputs, uint sampleFrames) = 0;
	virtual void downsample(float** outputs, uint sampleFrames) = 0;
	virtual void deleteResamplers() = 0;
	virtual void recreateResamplers() = 0;

	//
	// Helpers
protected:
	void nameHook(VstInt32 opcode, size_t maxLen, char* out);
	bool	doubleProcessingErrorWasTriggered;

	// 
	// GUI
#ifdef ENABLE_GUI
public:
	LRESULT wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

protected:
	void openWindow();
	void closeWindow();
	HWND	hostHwnd;
	HWND	selfHwnd;
	HWND	clientHwnd;
	ERect*	clientRect;
	ERect	selfRect;
#endif

	//
	// VST
public:
	void rebuildAeffect();
	AEffect				self;
	AEffect*			client;
	audioMasterCallback	audioMaster;


	//
	// Buffers
protected:
	void deleteChannels();
	void recreateChannels();
	bool		channelsValid;
	uint		numUpsamplers;
	uint		numDownsamplers;
	VstInt32	blockSize;
	float		sampleRate;

	float**		inputBuffers;
	float**		outputBuffers;
};

template <typename TUpsampler, typename TDownsampler> class WrapperProcess : public Wrapper
{
public:
	WrapperProcess();
	~WrapperProcess();

	virtual void upsample(float** inputs, uint sampleFrames);
	virtual void downsample(float** outputs, uint sampleFrames);
	virtual void deleteResamplers();
	virtual void recreateResamplers();

private:
	TUpsampler*		upsamplers;
	TDownsampler*	downsamplers;
};

typedef WrapperProcess< hiir::Upsampler2xSse<N_COEFFS_UP>, hiir::Downsampler2xSse<N_COEFFS_DOWN> > WrapperSse;
typedef WrapperProcess< hiir::Upsampler2x3dnow<N_COEFFS_UP>, hiir::Downsampler2x3dnow<N_COEFFS_DOWN> > Wrapper3dnow;
typedef WrapperProcess< hiir::Upsampler2xFpu<N_COEFFS_UP>, hiir::Downsampler2xFpu<N_COEFFS_DOWN> > WrapperFpu;

// ---------------------------------------------------------------------------
//
// globals
//


bool					g_isActive;
HINSTANCE				g_hInstance;
HMODULE					g_vstDll;
VstProc*				g_entry;
CRITICAL_SECTION		g_entryLock;
audioMasterCallback		g_audioMaster;
Wrapper*				g_currentWrapper;
int						g_cpuType;

// ---------------------------------------------------------------------------
//
// log helpers
//

#ifdef LOGGING_ENABLED
FILE* g_logFile;
void ____toLog(char* file, int lineNumber, char* format, ...)
{
	va_list args;

	if(!g_logFile) {
		return;
	}

	fprintf(g_logFile, "[%s:%d]\t", file, lineNumber);
    va_start(args, format);
	vfprintf(g_logFile, format, args);
    va_end(args);
	fprintf(g_logFile, "\r\n");
}
#endif // LOGGING_ENABLED

// ---------------------------------------------------------------------------
//
// aligned new/delete
//
// Technically, this is only needed for the SSE backend. However, it's easy
// to simply align everything, and there's also a minor speed improvement when
// data is aligned. 

inline void* _cdecl operator new(size_t cbSize)
{
	return _aligned_malloc(cbSize, 16);
}


inline void _cdecl operator delete(void* p)
{
	if(p) {
		_aligned_free(p);
	}
}


// ---------------------------------------------------------------------------
//
// Initialization and Breakdown
//

static void performCpuCheck()
{
	toLog("performCpuCheck");

	DWORD	extended = 0;
	DWORD	features = 0;

	_asm {
			push	edx

			// get regular feature bits
			mov		eax, 1
			cpuid
			mov		features, edx

			// detect if this is an AMD chip
			xor		edx, edx
			mov		eax, 0x80000000
			cpuid
			cmp		eax, 0x80000000
			jc		notAmd

			// it's an AMD chip, get extended bits
			mov		eax, 0x80000001
			cpuid
			mov		extended, edx

		notAmd:
			pop		edx
	}

	if(features & 0x02000000) {
		toLog("performCpuCheck() result: SSE");
		g_cpuType = CPUTYPE_SSE;
	} else if(extended & 0x80000000) {
		toLog("performCpuCheck() result: 3DNow!");
		g_cpuType = CPUTYPE_3DNOW;
	} else {
		toLog("performCpuCheck() result: x87");
		g_cpuType = CPUTYPE_X87;
	}
}


static void init()
{
	g_isActive = false;
	g_vstDll = NULL;

#ifdef ENABLE_GUI
	// register window classes
	WNDCLASSEX wcx; 
	ZeroMemory(&wcx, sizeof(WNDCLASSEX));
    wcx.cbSize = sizeof(wcx);
    wcx.style = CS_HREDRAW | CS_VREDRAW;
	wcx.lpfnWndProc = wndProc;
    wcx.hInstance = g_hInstance;
    wcx.lpszClassName = CLASSNAME_SELF;
	//wcx.hbrBackground = (HBRUSH) GetStockObject(LTGRAY_BRUSH);
	RegisterClassEx(&wcx);	// self

	wcx.lpfnWndProc = DefWindowProc;
    wcx.lpszClassName = CLASSNAME_CLIENT;
	RegisterClassEx(&wcx);	// client
#endif

	// :TODO: check for ERROR_CLASS_ALREADY_EXISTS

	// Get our path
	char	pathName[MAX_PATH_SIZE + 1];

	GetModuleFileName(g_hInstance, pathName, MAX_PATH_SIZE); // TODO: Check return value
	pathName[MAX_PATH_SIZE] = '\0'; // make sure the string is null-terminated

	size_t	len = strlen(pathName);
	char*	extension = pathName + (len - 7);

	if(strcmp(extension, ".os.dll") && strcmp(extension, "_os.dll")) {
		MessageBox(0, "Invalid .dll name - please consult Readme.txt", TITLE, MB_OK | MB_ICONERROR);
		return;
	}

#ifdef LOGGING_ENABLED
	// if needed, open log file
	strcpy(extension, "_os_log.txt");
	g_logFile = fopen(pathName, "w");

	toLog("Log initialized\n");
#endif // LOGGING_ENABLED

	// transform path
	strcpy(extension, ".dll");

	// Load
	toLog("Loading DLL: %s", pathName);
	g_vstDll = LoadLibrary(pathName);

	if(g_vstDll == NULL) {
		MessageBox(0, "Couldn't load associated .dll", TITLE, MB_OK | MB_ICONERROR);
		return;
	}

	// Get entry point
	g_entry = (VstProc*) GetProcAddress(g_vstDll, "VSTPluginMain");
	if(g_entry == NULL) {
		g_entry = (VstProc*) GetProcAddress(g_vstDll, "MAIN");
	}
	if(g_entry == NULL) {
		g_entry = (VstProc*) GetProcAddress(g_vstDll, "main");
	}
	if(g_entry == NULL) {
		MessageBox(0, "Associated .dll does not seem to be a valid VST Plugin", TITLE, MB_OK | MB_ICONERROR);
		return;
	}

	performCpuCheck();
	InitializeCriticalSection(&g_entryLock);
	g_isActive = true;
}


static void shutdown()
{
	if(g_vstDll != NULL) {
		FreeLibrary(g_vstDll);
	}

	DeleteCriticalSection(&g_entryLock);

#ifdef LOGGING_ENABLED
	toLog("Closing logfile");
	if(g_logFile) {
		fclose(g_logFile);
		g_logFile = NULL;
	}
#endif // LOGGING_ENABLED
}

/* :HACK:
 *
 * DllMain isn't supposed to call LoadLibrary or FreeLibrary. Unfortunately, it's
 * almost a necessity, as the alternative is all sorts of tricky, weird, and inefficient. 
 *
 * I'm guessing it won't cause any problems in the foreseeable future, as this .dll
 * loads no other .dlls except the VST it is supposed to host, which will then load
 * its dependencies as it needs.
 */
BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD fdwReason, LPVOID lpvReserved)
{
	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			g_hInstance = hInstance;
			init();
			return g_isActive ? TRUE : FALSE;

		case DLL_PROCESS_DETACH:
			shutdown();
			break;

		default:
			break;
	}

	return TRUE;
}

// ---------------------------------------------------------------------------
//
// Hooked VST dispatchers
//

VstIntPtr vst_audioMaster(AEffect* aeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	char blah[256];
	sprintf(blah, "enter with aeff %08X", aeff);
	Wrapper*			w;
	audioMasterCallback	audioMaster;

	if(aeff == NULL) {
		w = g_currentWrapper;
		audioMaster = g_audioMaster; 
	} else if(aeff->resvd1 == NULL) {
		w = g_currentWrapper;
		audioMaster = g_audioMaster;
	} else {
		w = (Wrapper*) aeff->resvd1;
		audioMaster = w->audioMaster;
	}

	w->audioMaster = audioMaster;

	VstIntPtr ret = w->vst_audioMaster(aeff, opcode, index, value, ptr, opt);
	return ret;
}

VstIntPtr vst_dispatcher(AEffect* aeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	return w->vst_dispatcher(opcode, index, value, ptr, opt);
}


void vst_processReplacing(AEffect* aeff, float** inputs, float** outputs, VstInt32 sampleFrames)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	w->vst_processReplacing(inputs, outputs, sampleFrames);
}


void vst_processAccumulating(AEffect* aeff, float** inputs, float** outputs, VstInt32 sampleFrames)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	w->vst_processAccumulating(inputs, outputs, sampleFrames);
}


void vst_processDoubleReplacing(AEffect* aeff, double** inputs, double** outputs, VstInt32 sampleFrames)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	w->vst_processDoubleReplacing(inputs, outputs, sampleFrames);
}


float vst_getParameter(AEffect* aeff, VstInt32 index)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	return w->vst_getParameter(index);
}


void vst_setParameter(AEffect* aeff, VstInt32 index, float value)
{
	Wrapper*	w = (Wrapper*) aeff->user;
	w->vst_setParameter(index, value);
}

// ---------------------------------------------------------------------------
//
// Entry point
//


AEffect* VSTPluginMain(audioMasterCallback audioMaster)
{
	toLog("VSTPluginMain");

	if(!g_isActive) {
		return NULL;
	}

	// :TRICKY:
	// A plugin may call the audioMaster before it has returned its
	// own AEffect. This is legal but makes our life difficult here.
	// We have to lock everything, save the current Wrapper and
	// audioMaster, then if the VST does call before the entry
	// returns, we can use the saved Wrapper. This means the host 
	// can't instantiate the plugin twice at the same time, but
	// hopefully it's not doing that anyway. ;)

	Wrapper*	w = Wrapper::createWrapper();

	toLog("Calling hosted plugin's entry");
	EnterCriticalSection(&g_entryLock); {
		
		w->audioMaster = g_audioMaster = audioMaster;
		g_currentWrapper = w;
		w->client = g_entry(vst_audioMaster);

	} LeaveCriticalSection(&g_entryLock);

	if(w->client == NULL) {
		toLog("No valid return from hosted plugin");
		delete w;
		return NULL;
	} else {
		toLog("Plugin returned seemingly valid AEffect");
	}

	w->rebuildAeffect();

	return &w->self;
}


// alternate entry points
AEffect* MAIN(audioMasterCallback audioMaster) 
{
	return VSTPluginMain(audioMaster); 
}

// ---------------------------------------------------------------------------
//
// Wrapper class
//

Wrapper::Wrapper()
{
	inputBuffers = NULL;
	outputBuffers = NULL;
	blockSize = 0;
	channelsValid = false;
	numDownsamplers = 0;
	numUpsamplers = 0;
	client = NULL;
#ifdef ENABLE_GUI
	selfHwnd = 0;
	hostHwnd = 0;
	clientHwnd = 0;
	clientRect = NULL;
#endif
	sampleRate = 0.0f;
	doubleProcessingErrorWasTriggered = false;
	memset(&self, 0, sizeof(AEffect));
}


Wrapper::~Wrapper()
{
	deleteChannels();
#ifdef ENABLE_GUI
	if(clientHwnd) {
		DestroyWindow(clientHwnd);
	}
	if(selfHwnd) {
		DestroyWindow(selfHwnd);
	}
#endif
}


void Wrapper::deleteChannels()
{
	if(numUpsamplers) {
		if(inputBuffers) {
			delete[] outputBuffers[0];
			delete[] inputBuffers;
		}
		inputBuffers = NULL;
		numUpsamplers = 0;
	}

	if(numDownsamplers) {
		if(outputBuffers) {
			delete[] outputBuffers[0];
			delete[] outputBuffers;
		}
		outputBuffers = NULL; 
		numDownsamplers = 0;
	}
}


void Wrapper::recreateChannels()
{
	bool	needReallocate = !channelsValid;

	if(client->numInputs > numUpsamplers) {
		needReallocate = true;
	}
	if(client->numOutputs > numDownsamplers) {
		needReallocate = true;
	}

	if(needReallocate) {
		deleteChannels();

		numUpsamplers = client->numInputs;
		numDownsamplers = client->numOutputs;

		inputBuffers = new float*[numUpsamplers];
		outputBuffers = new float*[numDownsamplers];

		if(numUpsamplers) {
			inputBuffers[0] = new float[blockSize * numUpsamplers];
			for(uint i = 1; i < numUpsamplers; i++) {
				inputBuffers[i] = inputBuffers[i - 1] + blockSize;
			}
		}
		if(numDownsamplers) {
			outputBuffers[0] = new float[blockSize * numDownsamplers];
			for(uint i = 1; i < numDownsamplers; i++) {
				outputBuffers[i] = outputBuffers[i - 1] + blockSize;
			}
		}

	}

	channelsValid = true;
} 

void Wrapper::rebuildAeffect()
{
	// usually a memcpy with soem special casing would be preferrable, and it's
	// what I did previously. However, some hosts (such as Chainer) seem to store
	// stuff in some weird places in the AEffect, so I changed it to use this 
	// method. This does mean, unfortunately, that future SDK updates could 
	// require an oversampler update (but since Steinborg is VST3-centered right
	// now anyway, ...

	// copy everything not special
	self.magic = client->magic;
	self.numPrograms = client->numPrograms;
	self.numParams = client->numParams;
	self.numInputs = client->numInputs;
	self.numOutputs = client->numOutputs;
	self.flags = client->flags;
	self.initialDelay = client->initialDelay;
	self.__realQualitiesDeprecated = client->__realQualitiesDeprecated;
	self.__offQualitiesDeprecated = client->__offQualitiesDeprecated;
	self.__ioRatioDeprecated = client->__ioRatioDeprecated;
	self.uniqueID = client->uniqueID;
	self.version = client->version;
	
	// callbacks
	self.processReplacing = ::vst_processReplacing;
	self.__processDeprecated = ::vst_processAccumulating;
	self.dispatcher = ::vst_dispatcher;
	self.getParameter = ::vst_getParameter;
	self.setParameter = ::vst_setParameter;
	self.processDoubleReplacing = ::vst_processDoubleReplacing;

	// prevent 64-bit processing
	self.flags &= ~effFlagsCanDoubleReplacing;

	// force processreplacing 
	self.flags |= effFlagsCanReplacing;

	// adjust latency
	self.initialDelay = client->initialDelay * 2;

	// pointers to the wrapper in the AEffects
	// resvd1 is reserved for the host, and since we're the host
	// of the client store it there. For our own AEffect, we save
	// in user, not object, because some hosts (REAPER comes to mind)
	// might try to do some AudioEffectX specific stuff if something is there. 
	client->resvd1 = (VstIntPtr) this;
	self.user = (void*) this;
}


Wrapper* Wrapper::createWrapper()
{
	Wrapper*	ret = NULL;

	switch(g_cpuType) {
		case CPUTYPE_SSE:
			ret = new WrapperSse();
			break;

		case CPUTYPE_3DNOW:
			ret = new Wrapper3dnow();
			break;
	
		case CPUTYPE_X87:
			ret = new WrapperFpu();
			break;

		default:
			MessageBox(0, "BUG: CPU type check wasn't performed!", TITLE, MB_OK | MB_ICONERROR);
			g_isActive = false;
			break;
	}

	return ret;	
}


// ---------------------------------------------------------------------------
//
// GUI
//

#ifdef ENABLE_GUI
extern "C" {
	LRESULT CALLBACK wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
		Wrapper* w = reinterpret_cast<Wrapper*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
		return w->wndProc(hWnd, msg, wParam, lParam);
	}
}


LRESULT Wrapper::wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
		case WM_PAINT:
			{
				PAINTSTRUCT	ps;
				HDC	hdc = BeginPaint(hWnd, &ps);
				RECT	r;
				r.top = r.left = 0;
				r.bottom = WND_OFFSET;
				r.right = clientRect->right;

				char	text[1024];
				sprintf(text, "%s (hosted plugin running at %.0fHz)", TITLE, sampleRate);
				DrawText(hdc, text, strlen(text), &r, 0);
				EndPaint(hWnd, &ps);
			}
			break;	// pass on to DefWindowProc as well
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}


void Wrapper::openWindow()
{
	if(!selfHwnd) {
		selfHwnd = CreateWindowEx(	0,
									CLASSNAME_SELF, 
									CLASSNAME_SELF, 
									WS_CHILD | WS_VISIBLE,
									0,
									0,
									clientRect->right,
									clientRect->bottom + WND_OFFSET,
									hostHwnd,
									0,
									g_hInstance,
									0);
		SetWindowLongPtr(selfHwnd, GWLP_USERDATA, (LONG_PTR) this);
	}

	if(!clientHwnd) {
		clientHwnd = CreateWindowEx(	0,
										CLASSNAME_CLIENT,
										CLASSNAME_CLIENT,
										WS_CHILD | WS_VISIBLE,
										0,
										WND_OFFSET,
										clientRect->right,
										clientRect->bottom,
										selfHwnd,
										0,
										g_hInstance,
										0);
	}

	ShowWindow(selfHwnd, SW_SHOWNORMAL);
	ShowWindow(clientHwnd, SW_SHOWNORMAL);
}


void Wrapper::closeWindow()
{
	CloseWindow(clientHwnd);
	clientHwnd = 0;
	CloseWindow(selfHwnd);
	selfHwnd = 0;
}
#endif


// ---------------------------------------------------------------------------
//
// dispatch helpers
//

void Wrapper::nameHook(VstInt32 opcode, size_t maxLen, char* out)
{
	char*	buffer = new char[maxLen + strlen(NAME_PREFIX) + 1];
	strcpy(buffer, NAME_PREFIX);
	client->dispatcher(client, opcode, 0, 0, buffer + strlen(NAME_PREFIX), 0.0f);
	buffer[maxLen - 1] = '\0';
	strcpy(out, buffer);	
	delete buffer;
}


void processEvents(VstEvents* events, int numerator, int denominator)
{
	for(VstInt32 i = 0; i < events->numEvents; i++) {
		VstEvent*	vstEvent = events->events[i];
		vstEvent->deltaFrames = vstEvent->deltaFrames * numerator / denominator;
	}
}


void getTimeInfo(VstTimeInfo* info)
{
	if(info) { 
		info->samplePos *= 2.0;
		info->sampleRate *= 2.0f;
		info->samplesToNextClock *= 2;
	}
}

// ---------------------------------------------------------------------------
//
// Hooked VST dispatchers
//

VstIntPtr Wrapper::vst_audioMaster(AEffect* originalAeff, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	toLog("audioMaster %d", opcode);
	VstIntPtr			ret;
	AEffect*			audioEffect;
	
	if(originalAeff) {
		if(this->client) {
			if(originalAeff != this->client) {
				char blah[256];
				sprintf(blah, "WTF!? originalAeff is %08X, this->client is %08X", originalAeff, this->client);
				MessageBoxA(0, blah, "Wrapper::vst_audioMaster", MB_OK);
			} 
			audioEffect = &self;
		} else {
			// this should, theoretically, never happen. It does, though (which sucks ass).
			// what we need to do here is work around this by setting it now
			this->client = originalAeff;
			this->rebuildAeffect();
			audioEffect = &self;
		}
	} else {
		audioEffect = NULL;
	}

	switch(opcode) {
		case audioMasterProcessEvents:
			processEvents((VstEvents*) ptr, 1, 2);
			break;

		case audioMasterIOChanged:
			channelsValid = false;
			break;
			
		default:
			break;
	}

	ret = audioMaster(audioEffect, opcode, index, value, ptr, opt);

	switch(opcode) {
		case audioMasterGetTime:
			getTimeInfo((VstTimeInfo*) ptr);
		default:
			break;
	}

	toLog("audioMaster end");
	return ret;
}

VstIntPtr Wrapper::vst_dispatcher(VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	toLog("dispatcher %d", opcode);
	bool		callDispatch = true;
	VstIntPtr	ret = 0;

	// pre-filter
	switch(opcode) {
		case effProcessEvents:
			processEvents((VstEvents*) ptr, 2, 1);
			break;

		case effMainsChanged:
			if(value != 0) {
				rebuildAeffect();
				recreateChannels();
				recreateResamplers();
			}
			break;

		case effSetSampleRate:
			opt *= 2.0f;	// report double sampleRate to plugin
			sampleRate = opt;
			break;

		case effSetBlockSize:
			value *= 2;		// report double blocksize to plugin
			if(value > blockSize) {
				channelsValid = false;
				blockSize = value;
			}
			break;

		case effGetEffectName:
			nameHook(effGetEffectName, kVstMaxEffectNameLen, (char*) ptr);
			callDispatch = false;
			break;

		case effGetProductString:
			nameHook(effGetProductString, kVstMaxProductStrLen, (char*) ptr);
			callDispatch = false;
			break;

#ifdef ENABLE_GUI
		case effEditOpen:
			hostHwnd = (HWND) ptr;
			openWindow();
			ptr = (void*) clientHwnd;
			break;

		case effEditClose:
			closeWindow();
			break;

		case effEditGetRect:
			{
				ERect**	hostRect = (ERect**) ptr;
				client->dispatcher(client, opcode, index, value, (void**) &clientRect, opt);
				memcpy(&selfRect, clientRect, sizeof(ERect));
				selfRect.bottom += WND_OFFSET;
				*hostRect = &selfRect;
				callDispatch = false;
			}			
#endif
	}

	// call client
	if(callDispatch) {
		toLog("dispatcher calling hosted plugin");
		ret = client->dispatcher(client, opcode, index, value, ptr, opt);
	}

	// post-filter
	switch(opcode) {
		case effClose:
			delete this;
			break;
	}

	toLog("dispatcher end");
	return ret;
}


//
// Processing
//

void Wrapper::vst_processReplacing(float** inputs, float** outputs, VstInt32 sampleFrames)
{
	upsample(inputs, sampleFrames);
	if(client->flags & effFlagsCanReplacing) {
		client->processReplacing(client, inputBuffers, outputBuffers, sampleFrames * 2);
	} else {
		SecureZeroMemory(outputBuffers[0], sizeof(float) * blockSize * numDownsamplers);
		client->__processDeprecated(client, inputBuffers, outputBuffers, sampleFrames * 2);
	}
	downsample(outputs, sampleFrames);
}


void Wrapper::vst_processAccumulating(float** inputs, float** outputs, VstInt32 sampleFrames)
{
	upsample(inputs, sampleFrames);

	// optimization: if it can processReplacing, do so instead of pre-zeroing the buffers
	if(client->flags & effFlagsCanReplacing) {
		client->processReplacing(client, inputBuffers, outputBuffers, sampleFrames * 2);
	} else {
		SecureZeroMemory(outputBuffers[0], sizeof(float) * blockSize * numDownsamplers);
		client->__processDeprecated(client, inputBuffers, outputBuffers, sampleFrames * 2);
	}

	// Downsample into the same buffers again
	downsample(outputBuffers, sampleFrames);

	// .. and now add to existing buffer
	for(uint i = 0; i < numDownsamplers; i++) {
		uint	n = sampleFrames;
		float*	s = outputBuffers[i];
		float*	d = outputs[i];
		while(n--) {
			*d += *s;
			d++; s++; 
		}
	}
}


void Wrapper::vst_processDoubleReplacing(double** inputs, double** outputs, VstInt32 sampleFrames)
{
	if(!doubleProcessingErrorWasTriggered) {
		MessageBox(0, "64-bit processing function was called - this is a Host bug!",
			TITLE, MB_OK | MB_ICONERROR);
		g_isActive = false;
	}
}


// 
// Parameters
//

float Wrapper::vst_getParameter(VstInt32 index)
{
	return client->getParameter(client, index);
}


void Wrapper::vst_setParameter(VstInt32 index, float value)
{
	client->setParameter(client, index, value);
}

// ---------------------------------------------------------------------------
//
// WrapperProcess class
//

template <typename TUpsampler, typename TDownsampler> 
WrapperProcess<TUpsampler, TDownsampler>::WrapperProcess()
{
	downsamplers = NULL;
	upsamplers = NULL;
}


template <typename TUpsampler, typename TDownsampler> 
WrapperProcess<TUpsampler, TDownsampler>::~WrapperProcess()
{
	deleteResamplers();
}


template <typename TUpsampler, typename TDownsampler> 
void WrapperProcess<TUpsampler, TDownsampler>::upsample(float** inputs, uint sampleFrames)
{
	for(int i = 0; i < client->numInputs; i++) {
		upsamplers[i].process_block(inputBuffers[i], inputs[i], sampleFrames);
	}
}


template <typename TUpsampler, typename TDownsampler> 
void WrapperProcess<TUpsampler, TDownsampler>::downsample(float** outputs, uint sampleFrames)
{
	for(int i = 0; i < client->numOutputs; i++) {
		downsamplers[i].process_block(outputs[i], outputBuffers[i], sampleFrames);
	}
}


template <typename TUpsampler, typename TDownsampler> 
void WrapperProcess<TUpsampler, TDownsampler>::recreateResamplers()
{
	deleteResamplers();
	upsamplers = new TUpsampler[numUpsamplers];
	downsamplers = new TDownsampler[numDownsamplers];
}


template <typename TUpsampler, typename TDownsampler> 
void WrapperProcess<TUpsampler, TDownsampler>::deleteResamplers()
{
	if(upsamplers) {
		delete[] upsamplers;
		upsamplers = NULL;
	}
	if(downsamplers) {
		delete[] downsamplers;
		downsamplers = NULL;
	}
}

